import unittest
import os
import threading
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops import mylogger, mqttclient, read_config
from eurydike.eventdetectionmanager import EventDetectionManager
from time import sleep


class _Tools:
    def _on_response_below(self, value):
        self.last_response_below = value.decode("utf-8")
        self.on_response_below_sync.set()

    def _on_response_outside(self, value):
        self.last_response_outside = value.decode("utf-8")
        self.on_response_outside_sync.set()

    def _on_value(self, value):
        self.last_value = value.decode("utf-8")
        self.on_value_sync.set()

    def _publish(self, topic, message):
        self.mqtt_client.publish(topic, message)
        if not self.on_value_sync.wait(1):
            self.fail("Timeout on publishing {}:{}.".format(topic, message))

        if isinstance(message, int):
            self.last_value = int(self.last_value)
        elif isinstance(message, float):
            self.last_value = float(self.last_value)
        else:
            self.fail("Dont know how to handle type '{}'.".format(type(message)))

        if message != self.last_value:
            self.fail("Last value '{}' not identical to last transmitted message '{}' on topic '{}'.".
                      format(self.last_value, message, topic))
        self.on_value_sync.clear()

    def _wait_for_response_outside(self, message):
        if not self.on_response_outside_sync.wait(1):
            self.fail("Timeout waiting on outside response for message '{}'.".format(message))
        self.on_response_outside_sync.clear()

    def _wait_for_response_below(self, message):
        if not self.on_response_below_sync.wait(1):
            self.fail("Timeout waiting on below response for message '{}'.".format(message))
        self.on_response_below_sync.clear()

    def _setUp(self):
        self.mqtt_client = mqttclient.MQTTClient(self.config["mqtt"], self.logger, True)

        self.mqtt_client.subscribe(self.topic_pub_below, self._on_response_below)
        self.mqtt_client.subscribe(self.topic_pub_outside, self._on_response_outside)
        self.mqtt_client.subscribe(self.topic_sub, self._on_value)
        self.mqtt_client.connect()

        self.on_value_sync = threading.Event()
        self.on_value_sync.clear()
        self.on_response_below_sync = threading.Event()
        self.on_response_below_sync.clear()
        self.on_response_outside_sync = threading.Event()
        self.on_response_outside_sync.clear()

        sleep(0.1)
        edm = None
        self.last_response_below = None
        self.last_response_outside = None
        self.last_value = None


class TestEventDetectionManager(unittest.TestCase, _Tools):
    @classmethod
    def setUpClass(cls):
        cls.config = read_config.read_config(
            os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config.yaml")
        cls.logger = mylogger.create_logger(cls.config["logger"], __name__)
        cls.logger.info("TestEventDetectionManager - start")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("TestEventDetectionManager - stop")

    def setUp(self):
        self.topic_sub = "/test/value"
        self.topic_pub_below = "/test/below"
        self.topic_pub_outside = "/test/band"
        self.on_violation = "event_detected"
        self.on_restoration = "event_ended"
        self._setUp()

    def tearDown(self):
        self.mqtt_client.disconnect()
        self.mqtt_client = None

    def test_edm_init(self):
        edm = EventDetectionManager(self.config, self.mqtt_client, self.logger)
        self.assertEqual(len(edm._event_detectors), 2)
        self.assertEqual(edm._event_detectors[0]._name, "below")
        self.assertEqual(edm._event_detectors[1]._name, "outside")

    def test_edm_start(self):
        edm = EventDetectionManager(self.config, self.mqtt_client, self.logger)
        edm.start()
        try:
            self.assertIn(edm._event_detectors[0]._message_handler,
                          edm._mqtt_client._topic_handler[self.topic_sub])
            self.assertIn(edm._event_detectors[1]._message_handler,
                          edm._mqtt_client._topic_handler[self.topic_sub])
        except KeyError:
            self.fail("No handler registered in mqtt client for topic '{}'.".format(self.topic_sub))

    def test_edm_steps(self):
        edm = EventDetectionManager(self.config, self.mqtt_client, self.logger)
        edm.start()
        ed_below = edm._event_detectors[0]
        ed_outside = edm._event_detectors[1]

        self._publish(self.topic_sub, 7.1)
        sleep(0.1)
        self.assertFalse(ed_below._event_active)
        self.assertIsNone(self.last_response_below)
        self.assertFalse(ed_outside._event_active)
        self.assertIsNone(self.last_response_outside)

        self._publish(self.topic_sub, 7.0)
        sleep(0.1)
        self.assertFalse(ed_below._event_active)
        self.assertIsNone(self.last_response_below)
        self.assertFalse(ed_outside._event_active)
        self.assertIsNone(self.last_response_outside)

        self._publish(self.topic_sub, 6.9)
        sleep(0.1)
        self._wait_for_response_outside(6.9)
        self._wait_for_response_below(6.9)
        self.assertTrue(ed_below._event_active)
        self.assertIsNotNone(self.last_response_below)
        self.assertTrue(ed_outside._event_active)
        self.assertIsNotNone(self.last_response_outside)
        self.assertEqual(self.last_response_below, self.on_violation)
        self.assertEqual(self.last_response_outside, self.on_violation)
        self.last_response_below = None
        self.last_response_outside = None

        self._publish(self.topic_sub, 8.1)
        self._wait_for_response_below(8.1)
        sleep(0.1)
        self.assertFalse(ed_below._event_active)
        self.assertIsNotNone(self.last_response_below)
        self.assertEqual(self.last_response_below, self.on_restoration)
        self.last_response_below = None
        self.assertTrue(ed_outside._event_active)
        self.assertIsNone(self.last_response_outside)

        self._publish(self.topic_sub, 7.2)
        self._wait_for_response_outside(7.2)
        sleep(0.1)
        self.assertFalse(ed_below._event_active)
        self.assertIsNone(self.last_response_below)
        self.assertFalse(ed_outside._event_active)
        self.assertIsNotNone(self.last_response_outside)
        self.assertEqual(self.last_response_outside, self.on_restoration)
        self.last_response_outside = None

    def test_edm_stop(self):
        edm = EventDetectionManager(self.config, self.mqtt_client, self.logger)
        edm.start()
        edm.stop()
        try:
            self.assertNotIn(edm._event_detectors[0]._message_handler,
                             edm._mqtt_client._topic_handler[self.topic_sub])
            self.assertNotIn(edm._event_detectors[1]._message_handler,
                             edm._mqtt_client._topic_handler[self.topic_sub])
        except KeyError:
            pass


if __name__ == '__main__':
    unittest.main()




from pelops import mylogger
from eurydike.eventdetectors.onband import OnBand
from eurydike.eventdetectors.onthreshold import OnThreshold


class EventDetectorFactory:
    """
    Create event detector instances based on the given yaml config structure.
    """

    @staticmethod
    def create_element(config_element, mqtt_client, parent_logger):
        """
        Create the corresponding eventdetection class instance.

        Note: new implementations of abstracteventdetector must be added to this method manually.

        :param config_element: config yaml structure for one eventdetector
        :param mqtt_client: mqtt client instance
        :param parent_logger: logger instance - a child will be spawned
        :return: a single eventdetection class instance
        """
        logger = mylogger.get_child(parent_logger, __name__)
        element = None
        if config_element["active"]:
            if config_element["type"].lower() == "onband":
                element = OnBand(config_element, mqtt_client, parent_logger)
            elif config_element["type"].lower() == "onthreshold":
                element = OnThreshold(config_element, mqtt_client, parent_logger)
            else:
                logger.error("EventDetectorFactory.create_element - unknown type '{}'".
                              format(config_element["type"].lower()))
                raise ValueError("Factory.create_element - unknown type '{}'".
                                 format(config_element["type"].lower()))
        else:
            logger.info("EventDetectorFactory.create_element - skipping inactive element '{}.{}'.".
                        format(config_element["type"].lower(), config_element["name"]))

        return element

    @staticmethod
    def create_elements(config_elements, mqtt_client, parent_logger):
        """
        Take a list of eventdetection config structures and create the corresponding class instances.

        :param config_elements: config yaml structure
        :param mqtt_client: mqtt client instance
        :param parent_logger: logger instance - a child will be spawned
        :return: a list with the eventdetector instances
        """
        logger = mylogger.get_child(parent_logger, __name__)
        element_list = []

        logger.info("EventDetectorFactory.create_elements - start")

        for config_element in config_elements:
            element = EventDetectorFactory.create_element(config_element, mqtt_client, parent_logger)
            if element is not None:
                element_list.append(element)

        logger.info("EventDetectorFactory.create_elements - finished")

        return element_list





from eurydike.eventdetectors.abstracteventdetector import AbstractEventDetector


class OnBand(AbstractEventDetector):
    """
    Each incoming value is checked against an upper and a lower threshold. The threshold is violated if
    the following statement is fulfilled: lower_threshold <= value <= upper_threshold.

        name: outside  # unqiue name for event detector
        type: onband  # detector type identifier
        upper-threshold: 8  # upper threshold for on band detection
        lower-threshold: 7  # lower threshold for on band detection
        topic-sub: /test/value
        topic-pub: /test/band
        responses:  # leave value empty or remove line for no response
          on-violation: event_detected  # on detection of a threshold violation send this value to topic-pub
          on-restoration: event_ended  # on the event of returning to valid values send this value to topic-pub
        active: True  # entry ignored if set to False
    """

    _upper_threshold = None  # upper threshold value
    _lower_threshold = None  # lower threshold value

    def __init__(self, config, mqtt_client, parent_logger):
        """
        Constructor

        :param config: config yaml structure
        :param mqtt_client: mqtt client instance
        :param parent_logger: parent logger - a child logger will be spawned
        """
        AbstractEventDetector.__init__(self, config, mqtt_client, parent_logger)
        self._upper_threshold = self._config["upper-threshold"]
        self._lower_threshold = self._config["lower-threshold"]
        if self._lower_threshold >= self._upper_threshold:
            self._logger.error("OnBand.__init__ ({}) - lower limit ({}) must be smaller than upper limit({}).".
                               format(self._name, self._lower_threshold, self._upper_threshold))
            raise ValueError("OnBand.__init__ ({}) - lower limit ({}) must be smaller than upper limit({}).".
                               format(self._name, self._lower_threshold, self._upper_threshold))

    def _is_violation(self, value):
        return not (self._lower_threshold <= value <= self._upper_threshold)



from pelops.abstractmicroservice import AbstractMicroservice
from eurydike.eventdetectors.eventdetectorfactory import EventDetectorFactory
import eurydike.schema as schema


class EventDetectionManager(AbstractMicroservice):
    """
    Creates, stores, starts, and stops all configured event detectors.

    mqtt:
        mqtt-address: localhost
        mqtt-port: 1883
        mqtt-credentials: ~/credentials.yaml
        log-level: INFO

    logger:
        log-level: DEBUG  # DEBUG, INFO, WARNING, ERROR, CRITICAL
        log-file: test_eurydike.log

    eventdetectors:
        - name: above  # unqiue name for event detector
          type: onthreshold  # detector type identifier
          comparator: gt  # GREATERTHAN/GT/>, LOWERTHAN/LT/<, EQUALTO/==
          threshold: 7  # threshold in combintation with comparator and value from topic-sub
          topic-sub: /test/value
          topic-pub: /test/above
          responses:  # leave value empty or remove line for no response
    #          on-violation: event_detected  # on detection of a threshold violation send this value to topic-pub
              on-restoration: event_ended  # on the event of returning to valid values send this value to topic-pub
          active: False  # entry ignored if set to False
    """

    _event_detectors = None  # list of event detector instances

    def __init__(self, config, mqtt_client=None, logger=None):
        """
        Constructor.

        :param config: config yaml structure
        :param mqtt_client: instance of an mqtt client (optional)
        :param logger: instance of a logger (optional)
        """
        AbstractMicroservice.__init__(self, config, "eventdetectors", mqtt_client=mqtt_client, logger=logger)
        self._event_detectors = EventDetectorFactory.create_elements(self._config, self._mqtt_client, self._logger)

    @classmethod
    def _get_schema(cls):
        return schema.get_schema()

    def _get_description(self):
        return "Eurydike is a simple event detection. Reacts to above-threshold, below-thershold, "\
                "and outside value-band."

    def _start(self):
        """
        start all event detectors
        """

        for event in self._event_detectors:
            event.start()

    def _stop(self):
        """
        stop all event detectors
        """
        for event in self._event_detectors:
            event.stop()



from pelops import mylogger


class AbstractEventDetector:
    """
    An event occurs when a value is outside the defined thresholds. Then the event detector changes its state
    to active and publishes the defined response. As soon as the value is back within the thresholds the state
    switches back to inactive and the defined response is published.
    """

    _config = None  # config yaml strucutre
    _logger = None  # logger instance
    _name = None  # name of event detector
    _mqtt_client = None  # mqtt client instance
    _event_active = None  # state of event detection - true if event has occured and still continues
    _topic_sub = None  # input for event detection
    _topic_pub = None  # publish results to this topic
    _responses = None  # stores the response messages that are published to _topic_pub

    def __init__(self, config, mqtt_client, parent_logger):
        """
        Constructor

        :param config: config yaml structure
        :param mqtt_client: mqtt client instance
        :param parent_logger: parent logger - a child logger will be spawned
        """
        self._config = config
        self._logger = mylogger.get_child(parent_logger, self.__class__.__name__, config)
        self._logger.info("{}.__init__ - initializing".format(self.__class__.__name__))
        self._logger.debug("{}.__init__ - configuration: {}".format(self.__class__.__name__, self._config))

        self._name = self._config["name"]

        self._mqtt_client = mqtt_client
        self._topic_pub = self._config["topic-pub"]
        self._topic_sub = self._config["topic-sub"]
        self._responses = {}
        for k, v in self._config["responses"].items():
            if v is None or v == "":
                continue
            self._responses[k] = v

        self._event_active = False

    def _message_handler(self, value):
        """
        The message handler is registered to process a message published to _topic_sub. Incoming values are
        checked if they are within the boundaries (_is_violation) and the corresponding response is published.

        :param value: incoming message from topic _topic_sub
        """
        try:
            value = int(value)
        except ValueError:
            value = float(value)

        if self._is_violation(value):
            # threshold is violated
            if not self._event_active:
                self._logger.info("{}._message_handler ({}) - threshold violation detected (value: {}).".
                                  format(self.__class__.__name__, self._name, value))
                try:
                    self._mqtt_client.publish(self._topic_pub, self._responses["on-violation"])
                except KeyError:
                    pass
            self._event_active = True
        else:
            if self._event_active:
                self._logger.info("{}._message_handler ({}) - threshold restored (value: {}).".
                                  format(self.__class__.__name__, self._name, value, ))
                try:
                    self._mqtt_client.publish(self._topic_pub, self._responses["on-restoration"])
                except KeyError:
                    pass
            self._event_active = False

    def _is_violation(self, value):
        """
        Checks if the value is within the boundaries/thresholds. Must be implemented by child classes.

        :param value: incoming value
        :return: True if the value violates the boundaries/thresholds. False if not.
        """
        raise NotImplementedError

    def start(self):
        """
        start event detection by subscribing to _topic_sub
        """
        self._mqtt_client.subscribe(self._topic_sub, self._message_handler)

    def stop(self):
        """
        stop event detection by unsubscribing from _topic_sub
        """
        self._mqtt_client.unsubscribe(self._topic_sub, self._message_handler)



def get_schema():
    return {
            "eventdetectors": {
                "description": "Root node for alcathous specific entries.",
                "type": "array",
                "items": {
                    "oneOf": [
                        {
                            "description": "on threshold",
                            "type": "object",
                            "properties": {
                                "name": {
                                    "description": "unqiue name for event detector",
                                    "type": "string"
                                },
                                "type": {
                                    "description": "detector type identifier",
                                    "type": "string",
                                    "enum": ["onthreshold"]
                                },
                                "comparator": {
                                    "description": "GREATERTHAN/GT/>, LOWERTHAN/LT/<, EQUALTO/==",
                                    "type": "string",
                                    "enum": ["GREATERTHAN", "GT", ">", "LOWERTHAN", "LT", "<", "EQUALTO", "==",
                                             "greaterthan", "gt", "lowerthan", "lt", "equalto"]
                                },
                                "threshold": {
                                    "description": "threshold in combintation with comparator and value from topic-sub",
                                    "type": "number"
                                },
                                "topic-pub": {
                                    "type": "string"
                                },
                                "topic-sub": {
                                    "type": "string"
                                },
                                "active": {
                                    "description": "entry ignored if set to False",
                                    "type": "boolean"
                                },
                                "responses": {
                                    "description": "leave value empty or remove line for no response",
                                    "type": "object",
                                    "properties": {
                                        "on-violation": {
                                            "description": "on detection of a threshold violation send this value to "
                                                           "topic-pub",
                                            "type": ["string", "null"]
                                        },
                                        "on-restoration": {
                                            "description": "on the event of returning to valid values send this value "
                                                           "to topic-pub",
                                            "type": ["string", "null"]
                                        }
                                    },
                                    "additionalProperties": False
                                }
                            },
                            "required": ["type", "comparator", "threshold", "topic-sub", "topic-pub", "active"],
                            "additionalProperties": False
                        },
                        {
                            "description": "on band",
                            "type": "object",
                            "properties": {
                                "name": {
                                    "description": "unqiue name for event detector",
                                    "type": "string"
                                },
                                "type": {
                                    "description": "detector type identifier",
                                    "type": "string",
                                    "enum": ["onband"]
                                },
                                "upper-threshold": {
                                    "description": "upper threshold for on band detection",
                                    "type": "number"
                                },
                                "lower-threshold": {
                                    "description": "lower threshold for on band detection",
                                    "type": "number"
                                },
                                "topic-pub": {
                                    "type": "string"
                                },
                                "topic-sub": {
                                    "type": "string"
                                },
                                "active": {
                                    "description": "entry ignored if set to False",
                                    "type": "boolean"
                                },
                                "responses": {
                                    "description": "leave value empty or remove line for no response",
                                    "type": "object",
                                    "properties": {
                                        "on-violation": {
                                            "description": "on detection of a threshold violation send this value to "
                                                           "topic-pub",
                                            "type": "string"
                                        },
                                        "on-restoration": {
                                            "description": "on the event of returning to valid values send this value "
                                                           "to topic-pub",
                                            "type": "string"
                                        }
                                    },
                                    "additionalProperties": False
                                }
                            },
                            "required": ["type", "lower-threshold", "upper-threshold", "topic-sub", "topic-pub",
                                         "active"],
                            "additionalProperties": False
                        }
                    ]
                },
                "additionalItems": False
            }
           }




import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.read_config import read_config, validate_config
from eurydike.eventdetectionmanager import EventDetectionManager


class TestValidateConfig(unittest.TestCase):
    def test_validate_config(self):
        config = read_config("config.yaml")
        validation_result = validate_config(config, EventDetectionManager._get_schema())
        self.assertIsNone(validation_result)


if __name__ == '__main__':
    unittest.main()



from eurydike.eventdetectors.abstracteventdetector import AbstractEventDetector
from enum import Enum


class Comparator(Enum):
    """
    The Comparators are used for OnThreshold checks. Each enum is represented by a lambda function.
    """

    GREATERTHAN = (lambda x, y: x > y),
    LOWERTHAN = (lambda x, y: x < y),
    EQUALTO = (lambda x, y: x == y),

    @classmethod
    def get_enum(cls, text, logger):
        """
        Takes a string and returns the corresponding enum.

        :param text: string from config yaml structure.
        :param logger: logger instance
        :return: enum
        :raises: ValueError: raised if the text cannot be matched to one of the enum.
        """
        if text.lower()=="greaterthan" or text.lower()=="gt" or text.lower()==">":
            return cls.GREATERTHAN
        elif text.lower()=="lowerthan" or text.lower()=="lt" or text.lower()=="<":
            return cls.LOWERTHAN
        elif text.lower()=="equalto" or text.lower()=="==":
            return cls.EQUALTO
        else:
            logger.error("Comperator.get_enum - unkown value '{}'.".format(text))
            raise ValueError("Comperator.get_enum - unkown value '{}'.".format(text))

    def __call__(self, *args, **kwargs):
        """
        Execute the lambda function
        :return: result of the lambda function
        """
        return self.value[0](*args, **kwargs)


class OnThreshold(AbstractEventDetector):
    """
    Checks for simple threshold violations. Can be <, >, or ==.

        name: above  # unqiue name for event detector
        type: onthreshold  # detector type identifier
        comparator: gt  # GREATERTHAN/GT/>, LOWERTHAN/LT/<, EQUALTO/==
        threshold: 7  # threshold in combintation with comparator and value from topic-sub
        topic-sub: /test/value
        topic-pub: /test/above
        responses:  # leave value empty or remove line for no response
          on-violation: event_detected  # on detection of a threshold violation send this value to topic-pub
          on-restoration: event_ended  # on the event of returning to valid values send this value to topic-pub
        active: False  # entry ignored if set to False
    """
    _threshold = None  # threshold value
    _comparator = None  # enum representing the comparator type (<, >, or ==)

    def __init__(self, config, mqtt_client, parent_logger):
        """
        Constructor

        :param config: config yaml structure
        :param mqtt_client: mqtt client instance
        :param parent_logger: parent logger - a child logger will be spawned
        """
        AbstractEventDetector.__init__(self, config, mqtt_client, parent_logger)
        self._comparator = Comparator.get_enum(self._config["comparator"], self._logger)
        self._threshold = self._config["threshold"]

    def _is_violation(self, value):
        return self._comparator(value, self._threshold)





import unittest
import threading
from time import sleep
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops import mylogger, mqttclient, read_config
from eurydike.eventdetectors.onthreshold import OnThreshold, Comparator


class _Tools:
    def _on_response(self, value):
        self.last_message = value.decode("utf-8")
        self.on_response_sync.set()

    def _on_value(self, value):
        self.last_value = value.decode("utf-8")
        self.on_value_sync.set()

    def _publish(self, topic, message):
        self.mqtt_client.publish(topic, message)
        if not self.on_value_sync.wait(1):
            self.fail("Timeout on publishing {}:{}.".format(topic, message))

        if isinstance(message, int):
            self.last_value = int(self.last_value)
        elif isinstance(message, float):
            self.last_value = float(self.last_value)
        else:
            self.fail("Dont know how to handle type '{}'.".format(type(message)))

        if message != self.last_value:
            self.fail("Last value '{}' not identical to last transmitted message '{}' on topic '{}'.".
                      format(self.last_value, message, topic))
        self.on_value_sync.clear()

    def _wait_for_response(self, message):
        if not self.on_response_sync.wait(1):
            self.fail("Timeout waiting on response for message '{}'.".format(message))
        self.on_response_sync.clear()

    def _setUp(self, id):
        conf = read_config.read_config(
            os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config.yaml")
        self.config = conf["eventdetectors"][id]
        self.logger = mylogger.create_logger(conf["logger"], __name__)
        self.mqtt_client = mqttclient.MQTTClient(conf["mqtt"], self.logger, True)

        self.mqtt_client.subscribe(self.topic_pub, self._on_response)
        self.mqtt_client.subscribe(self.topic_sub, self._on_value)
        self.mqtt_client.connect()

        self.on_value_sync = threading.Event()
        self.on_response_sync = threading.Event()
        self.on_value_sync.clear()
        self.on_response_sync.clear()

        self.ed = None
        self.last_message = None
        self.last_value = None


class TestOnThresholdBelow(unittest.TestCase, _Tools):
    def setUp(self):
        self.topic_sub = "/test/value"
        self.topic_pub = "/test/below"
        self.on_violation = "event_detected"
        self.on_restoration = "event_ended"
        self._setUp(1)

    def tearDown(self):
        self.mqtt_client.disconnect()

    def test_below_init(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.assertEqual(self.ed._comparator, Comparator.LOWERTHAN)
        self.assertEqual(self.ed._threshold, 7.0)
        self.assertEqual(self.ed._topic_sub, self.topic_sub)
        self.assertEqual(self.ed._topic_pub, self.topic_pub)
        self.assertEqual(self.ed._responses["on-violation"], self.on_violation)
        self.assertEqual(self.ed._responses["on-restoration"], self.on_restoration)
        self.assertFalse(self.ed._event_active)

    def test_below_start(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.ed.start()
        try:
            temp = self.mqtt_client._topic_handler[self.topic_sub]
            self.assertIn(self.ed._message_handler, temp)
        except KeyError:
            self.fail("No handler registered in mqtt client for topic '{}'.".format(self.topic_sub))

    def test_below_steps(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.ed.start()

        self._publish(self.topic_sub, 7.2)
        sleep(0.1)
        self.assertFalse(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 7.1)
        sleep(0.1)
        self.assertFalse(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 7)
        sleep(0.1)
        self.assertFalse(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 6.9)
        self._wait_for_response(6.9)
        self.assertTrue(self.ed._event_active)
        self.assertIsNotNone(self.last_message)
        self.assertEqual(self.last_message, self.on_violation)
        self.last_message = None

        self._publish(self.topic_sub, 6.8)
        sleep(0.1)
        self.assertTrue(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 7)
        self._wait_for_response(7)
        self.assertFalse(self.ed._event_active)
        self.assertIsNotNone(self.last_message)
        self.assertEqual(self.last_message, self.on_restoration)
        self.last_message = None

        self._publish(self.topic_sub, 7)
        sleep(0.1)
        self.assertFalse(self.ed._event_active)
        self.assertIsNone(self.last_message)

    def test_below_stop(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.ed.start()
        self.ed.stop()
        try:
            temp = self.mqtt_client._topic_handler[self.topic_sub]
            self.assertNotIn(self.ed._message_handler, temp)
        except KeyError:
            pass


class TestOnThresholdAbove(unittest.TestCase, _Tools):
    def setUp(self):
        self.topic_sub = "/test/value"
        self.topic_pub = "/test/above"
        self.on_violation = "event_detected"
        self.on_restoration = "event_ended"
        self._setUp(0)

    def tearDown(self):
        self.mqtt_client.disconnect()

    def test_above_init(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.assertEqual(self.ed._comparator, Comparator.GREATERTHAN)
        self.assertEqual(self.ed._threshold, 7.0)
        self.assertEqual(self.ed._topic_sub, self.topic_sub)
        self.assertEqual(self.ed._topic_pub, self.topic_pub)
        self.assertNotIn("on-violation", self.ed._responses)
        self.assertEqual(self.ed._responses["on-restoration"], self.on_restoration)
        self.assertFalse(self.ed._event_active)

    def test_above_start(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.ed.start()
        try:
            temp = self.mqtt_client._topic_handler[self.topic_sub]
            self.assertIn(self.ed._message_handler, temp)
        except KeyError:
            self.fail("No handler registered in mqtt client for topic '{}'.".format(self.topic_sub))

    def test_above_steps(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.ed.start()

        self._publish(self.topic_sub, 6.8)
        sleep(0.1)
        self.assertFalse(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 6.9)
        sleep(0.1)
        self.assertFalse(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 7)
        sleep(0.1)
        self.assertFalse(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 7.1)
        sleep(0.1)
        self.assertTrue(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 7.2)
        sleep(0.1)
        self.assertTrue(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 7)
        self._wait_for_response(7)
        self.assertFalse(self.ed._event_active)
        self.assertIsNotNone(self.last_message)
        self.assertEqual(self.last_message, self.on_restoration)
        self.last_message = None

        self._publish(self.topic_sub, 7)
        sleep(0.1)
        self.assertFalse(self.ed._event_active)
        self.assertIsNone(self.last_message)

    def test_above_stop(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.ed.start()
        self.ed.stop()
        try:
            temp = self.mqtt_client._topic_handler[self.topic_sub]
            self.assertNotIn(self.ed._message_handler, temp)
        except KeyError:
            pass


class TestOnThresholdEqual(unittest.TestCase, _Tools):
    def setUp(self):
        self.topic_sub = "/test/value"
        self.topic_pub = "/test/equal"
        self.on_violation = "event_detected"
        self.on_restoration = "event_ended"
        self._setUp(2)

    def tearDown(self):
        self.mqtt_client.disconnect()

    def test_equal_init(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.assertEqual(self.ed._comparator, Comparator.EQUALTO)
        self.assertEqual(self.ed._threshold, 7.0)
        self.assertEqual(self.ed._topic_sub, self.topic_sub)
        self.assertEqual(self.ed._topic_pub, self.topic_pub)
        self.assertEqual(self.ed._responses["on-violation"], self.on_violation)
        self.assertNotIn("on-restoration", self.ed._responses)
        self.assertFalse(self.ed._event_active)

    def test_equal_start(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.ed.start()
        try:
            temp = self.mqtt_client._topic_handler[self.topic_sub]
            self.assertIn(self.ed._message_handler, temp)
        except KeyError:
            self.fail("No handler registered in mqtt client for topic '{}'.".format(self.topic_sub))

    def test_equal_steps(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.ed.start()

        self._publish(self.topic_sub, 6.9)
        sleep(0.1)
        self.assertFalse(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 7.0)
        self._wait_for_response(7.0)
        self.assertTrue(self.ed._event_active)
        self.assertIsNotNone(self.last_message)
        self.assertEqual(self.last_message, self.on_violation)
        self.last_message = None

        self._publish(self.topic_sub, 7.0)
        sleep(0.1)
        self.assertTrue(self.ed._event_active)
        self.assertIsNone(self.last_message)

        self._publish(self.topic_sub, 7.1)
        sleep(0.1)
        self.assertFalse(self.ed._event_active)
        self.assertIsNone(self.last_message)

    def test_equal_stop(self):
        self.ed = OnThreshold(self.config, self.logger, self.mqtt_client)
        self.ed.start()
        self.ed.stop()
        try:
            temp = self.mqtt_client._topic_handler[self.topic_sub]
            self.assertNotIn(self.ed._message_handler, temp)
        except KeyError:
            pass


if __name__ == '__main__':
    unittest.main()
